#include <iostream>
#include <cxcore.h>
#include <cv.h>
#include <highgui.h>
#include <jpeglib.h>
#include <errno.h>
#include <highgui.h>
#include "cvblob.h"
#include <math.h>
#include "SerialComm.h"

#define PI 3.1415

const double BLACK_THRESHOLD = 85.0;
static unsigned char jpegQuality = 70;

// serial stuff
SerialComm *comm;
int err=0;
int num_bytes=0;
char command[20];
char version[100];
speed_t baud_rate = B230400;
char *port_name = (char*)"/dev/ttyS0\0";

using namespace cv;

/**
	Print error message and terminate programm with EXIT_FAILURE return code.

	\param s error message to print
*/
static void errno_exit(const char* s)
{
	fprintf(stderr, "%s error %d, %s\n", s, errno, strerror (errno));
	exit(EXIT_FAILURE);
}

/**
	Write image to jpeg file.

	\param img image to write
*/
static void jpegWrite(unsigned char* inputImg, int channels, char* name, int w, int h)
{

	unsigned char *img = (unsigned char*)malloc(w*h*channels);
	memcpy(img, inputImg, w*h*channels);

	if(channels == 3) {
		int i, x, y;
		unsigned char red, blue;
		i=0;
		for(y=0; y < h; y++)
		{
		    for(x=0; x < w; x++)
		    {
		        blue = (int)img[i];
		        red = (int)img[i+2];
			img[i] = red;
			img[i+2] = blue;
		        i+=3;
		    }
		}
	}

	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	
	JSAMPROW row_pointer[1];
	FILE *outfile = fopen( name, "wb" );

	// try to open file for saving
	if (!outfile) {
		errno_exit("jpeg");
	}

	// create jpeg data
	cinfo.err = jpeg_std_error( &jerr );
	jpeg_create_compress(&cinfo);
	jpeg_stdio_dest(&cinfo, outfile);

	// set image parameters
	cinfo.image_width = w;	
	cinfo.image_height = h;
	switch(channels) {
		case 3:
			cinfo.input_components = 3;
			cinfo.in_color_space = JCS_RGB;
			break;
		case 1: 
			cinfo.input_components = 1;
			cinfo.in_color_space = JCS_GRAYSCALE;
			break;
	}

	// set jpeg compression parameters to default
	jpeg_set_defaults(&cinfo);
	// and then adjust quality setting
	jpeg_set_quality(&cinfo, jpegQuality, TRUE);

	// start compress 
	jpeg_start_compress(&cinfo, TRUE);

	// feed data
	while (cinfo.next_scanline < cinfo.image_height) {
		row_pointer[0] = &img[cinfo.next_scanline * cinfo.image_width *  cinfo.input_components];
		jpeg_write_scanlines(&cinfo, row_pointer, 1);
	}

	// finish compression
	jpeg_finish_compress(&cinfo);

	// destroy jpeg data
	jpeg_destroy_compress(&cinfo);

	// close output file
	fclose(outfile);

	free(img);
}


int main(int argc, char* argv[]) {
	float angle = 0;
	signed int speedFactor=0;

    CvCapture*    camera = 0;
    //if (argc > 1)
    //    camera = cvCreateFileCapture(argv[1]);
    //else
    //    camera = cvCreateCameraCapture(1); // Use the default camera
    //    //camera = cvCaptureFromCAM(-1);


	camera = cvCreateCameraCapture(1);

	unsigned int hMin = atoi(argv[1]);
	unsigned int hMax = atoi(argv[2]);
	unsigned int sMin = atoi(argv[3]);
	unsigned int sMax = atoi(argv[4]);
	unsigned int vMin = atoi(argv[5]);
	unsigned int vMax = atoi(argv[6]);
	unsigned int debug = atoi(argv[7]);

	std::cout << "hMin: " << hMin << ", hMax: " << hMax << std::endl;
	std::cout << "sMin: " << sMin << ", sMax: " << sMax << std::endl;
	std::cout << "vMin: " << vMin << ", vMax: " << vMax << std::endl;


    int i=0;

    std::cerr << "camera opened" << std::endl;
    std::cerr << "frame width = " << (int)cvGetCaptureProperty(camera,CV_CAP_PROP_FRAME_WIDTH) << std::endl;
    std::cerr << "frame height = " << (int)cvGetCaptureProperty(camera,CV_CAP_PROP_FRAME_HEIGHT) << std::endl;
    std::cerr << "fps = " << (int)cvGetCaptureProperty(camera,CV_CAP_PROP_FPS) << std::endl;
    std::cerr << "format = " << (int)cvGetCaptureProperty(camera,CV_CAP_PROP_FORMAT) << std::endl;

    //if(cvSetCaptureProperty(camera, CV_CAP_PROP_FRAME_WIDTH, 352))
    //    printf("Cant set width property\n");

    //if(cvSetCaptureProperty(camera, CV_CAP_PROP_FRAME_HEIGHT, 288))
    //    printf("Cant set height property\n");

    //std::cerr << "frame width = " << (int)cvGetCaptureProperty(camera,CV_CAP_PROP_FRAME_WIDTH) << std::endl;
    //std::cerr << "frame height = " << (int)cvGetCaptureProperty(camera,CV_CAP_PROP_FRAME_HEIGHT) << std::endl;


    int keyPressed = 0;

	IplImage *image, *frame, *frameHSV, *segmentated, *labelImg;
	int picWidth, picHeight;
 
	CvTracks tracks;
	CvBlobs blobs;
	CvBlob* blob;

	CvSize imgSize;

	 unsigned int result = 0;


	// serial communication init
    	comm = new SerialComm();
	err = comm->connect(port_name, baud_rate);
	if(err == -1) {
        	perror("Cannot open serial port: ");
        	return 1;
    	}

	image = cvQueryFrame(camera); 

	imgSize = cvGetSize(image);

        frame=cvCreateImage(imgSize,image->depth,image->nChannels);
	frameHSV = cvCreateImage(imgSize, frame->depth, frame->nChannels);
	segmentated = cvCreateImage(imgSize, 8, 1);	// 8 bits, 1 channel
	labelImg = cvCreateImage(cvGetSize(frame), IPL_DEPTH_LABEL, 1);

    while(1) {

        for(i=0; i<5; i++) {
            cvQueryFrame(camera);//Need to get updated image
        }

        image = cvQueryFrame(camera); //Be careful not to change the value or to free the "frame" variable
        if(!frame) {
            std::cerr << "frame not grabbed properly" << std::endl;
            return -1;
        }

	//jpegWrite((unsigned char*)image->imageData, image->nChannels, (char*)"image.jpg", image->width, image->height);

	/*        
	std::cerr << "frame grabbed" << std::endl;
        std::cerr << "nChannels = " << image->nChannels << std::endl;
        std::cerr << "depth = " << image->depth << std::endl;
        std::cerr << "colorModel = " << image->colorModel << std::endl;
        std::cerr << "channelSeq = " << image->channelSeq << std::endl;
	*/

 	cvConvertScale(image, frame, 1, 0);	// copy image to frame
  
	cvCvtColor(frame, frameHSV, CV_BGR2HSV);

	cvInRangeS(frameHSV, cvScalar(hMin, sMin, vMin), cvScalar(hMax, sMax, vMax), segmentated);
	//cvInRangeS(frameHSV, cvScalar(120, 50, 50), cvScalar(180, 255, 255), segmentated);	// pink
 
	if(debug==1) {
		//jpegWrite((unsigned char*)segmentated->imageData, segmentated->nChannels, (char*)"segmented.jpg", segmentated->width, segmentated->height);
	}

	//Can experiment either or both
	cvSmooth(segmentated, segmentated, CV_MEDIAN, 7, 7);
	cvSmooth(segmentated, segmentated, CV_GAUSSIAN, 9, 9);

	if(debug==1) {
		jpegWrite((unsigned char*)segmentated->imageData, segmentated->nChannels, (char*)"segmented-smooth.jpg", segmentated->width, segmentated->height);
	}

 
	result = cvLabel(segmentated, labelImg, blobs);
	cvFilterByArea(blobs, 0, 1000000);
	cvRenderBlobs(labelImg, blobs, frame, frame, 0x000f, 1.);
	//cvUpdateTracks(blobs, tracks, 200., 5);
	//cvRenderTracks(tracks, frame, frame, 0x000f, NULL);
 
	//std::cerr << "Found " << blobs.size() << " blobs!" << std::endl;

	picWidth = frame->width;
	picHeight = frame->height;	


	if (cvGreaterBlob(blobs)) {
		blob = blobs[cvGreaterBlob(blobs)];
 		if(debug==1) {
			//std::cout << "Blobs found: " << blobs.size() << std::endl;
			//std::cout << "Pixels labeled: " << result << std::endl;
			std::cout << "center-x: " << blob->centroid.x << " center-y: " << blob->centroid.y << std::endl;
			//std::cout << "offset-x: " << ((picWidth / 2)-(blob->centroid.x)) << " offset-y: " << (picHeight / 2)-(blob->centroid.y) << std::endl;
			//std::cout << "\n";
		}

		angle = atan2(blob->centroid.y-240.0, blob->centroid.x-320.0)/PI*180.0;
		if(angle < 0) {
			angle += 360;
		}

		if(debug==1) {
			std::cout << "angle = " << angle << std::endl;
			std::cout << "l: " << 180-(int)angle << ", r: " << -(180-(int)angle) << std::endl;		
		}

		if(abs(180-(int)angle) > 10) {
			speedFactor = (180-(int)angle);
			// add base speed
			if(speedFactor > 0) {	
				speedFactor += 50;
			} else {
				speedFactor -= 50;
			}
			memset(command, 0x0, 20);
            		sprintf(command, "D,%d,%d\r\n", speedFactor, -speedFactor);
            		num_bytes = comm->writeData((char *)command, strlen(command), 6000);
		} else {
			memset(command, 0x0, 20);
            		sprintf(command, "D,%d,%d\r\n", 0, 0);
            		num_bytes = comm->writeData((char *)command, strlen(command), 6000);
		}

	} else {
		// stop if no blobs detected
		memset(command, 0x0, 20);
            	sprintf(command, "D,%d,%d\r\n", 0, 0);
            	num_bytes = comm->writeData((char *)command, strlen(command), 6000);
	}

	if(debug==1) {
		jpegWrite((unsigned char*)frame->imageData, frame->nChannels, (char*)"blob.jpg", frame->width, frame->height);
	}

	/*        
	std::cout << "Press 0 to stop";
        std::cin >> keyPressed;
        std::cout << std::endl;

        if (keyPressed == 0)
          break;
	*/
    }

    cvReleaseCapture(&camera);

	comm->disconnect();

    std::cerr << "camera released" << std::endl;

    return 0;
}
